"""This module contains some functions which may be useful to deal with 
triangular meshes, spherical functions and their visualization with matplotlib.
"""
from matplotlib import pyplot as plt
import numpy as np
import trimesh
from scipy.special import sph_harm_y


def view_mesh(mesh, ax, **kwargs):
    """Visualize a triangular mesh using matplotlib. Keyword arguments **kwargs
    are passed to matplotlib's plot_trisurf.

    Parameters
    ----------
    mesh : a trimesh.Trimesh instance
    ax : an Axes3D instance
    """
    ax.plot_trisurf(*mesh.vertices.T, triangles=mesh.faces, **kwargs)
    ax.axis("off")


def mesh_to_silhouette(mesh, sphere_points):
    """Transforms a mesh onto a silhouette.

    Parameters
    ----------
    mesh : trimesh.Trimesh instances
        Object to be transformed onto a spherical function.
    sphere_points : array-like (nb_points, 3) 

    Returns
    -------
    values : array-like, shape (nb_points, )
    """
    nb_points = sphere_points.shape[0]
    values = np.zeros(nb_points)
    M = mesh.centroid
    for i in range(nb_points):
        u = sphere_points[i]
        intersections, _, _ = mesh.ray.intersects_location(
            M[np.newaxis, :], u[np.newaxis, :])
        distances = np.linalg.norm(intersections - M[np.newaxis, :], axis=1)
        values[i] = np.max(distances)
    return values


def view_spherical_function(sphere, values, ax, **kwargs):
    """Represents a spherical function: provided a triangular mesh of a sphere,
    we displace the vertices of the mesh away from the center by the value of 
    the function in this direction. Keyword arguments **kwargs are passed to 
    matplotlib's plot_trisurf.
    
    Parameters
    ----------
    sphere : a trimesh.Trimesh instance
    values : array-like, shape (nb_points, )
    ax : Axes3D instance
    """
    points = np.copy(sphere.vertices)
    points *= np.abs(values[:, np.newaxis])
    deformed_sphere = trimesh.Trimesh(vertices=points, faces=sphere.faces)
    view_mesh(deformed_sphere, ax, **kwargs)


def spherical_harmonic(j, sphere):
    """Computes the spherical harmonic function of index j on the vertices of 
    the provided sphere.

    Parameters
    ----------
    j : int
    sphere : a trimesh.Trimesh instance
    """
    # we first convert into spherical coordinates
    theta = np.arccos(sphere.vertices[:, 2])
    phi = np.arctan2(sphere.vertices[:, 1], sphere.vertices[:, 0])

    l = int(np.sqrt(j))
    m = j - l**2 - l
    complex_values = sph_harm_y(l, abs(m), theta, phi)
    if m < 0:
        return np.sqrt(2) * (-1)**m * complex_values.imag
    elif m > 0:
        return np.sqrt(2) * (-1)**m * complex_values.real
    else: # m == 0
        return complex_values.real


